# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.
from __future__ import absolute_import, division, print_function

__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]

__title__ = "pipfile"
__summary__ = ""
__uri__ = "https://github.com/pypa/pipfile"

__version__ = "0.0.2"

__author__ = "Kenneth Reitz and individual contributors"
__email__ = "me@kennethreitz.org"

__license__ = "BSD or Apache License, Version 2.0"
__copyright__ = "Copyright 2017 %s" % __author__
