



set CI=1
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -c "import pipenv"
IF %ERRORLEVEL% NEQ 0 exit /B 1
pipenv --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
pip check
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
